/*
 *  Image to Bitmap converter functions.
 *
 *  This file implements conversion from a platform-independent
 *  Image data format into the X-Windows Bitmap data format.
 *
 *  Platform: X-Windows.
 *
 *  Version: 3.00  2001/05/05  First release.
 *  Version: 3.43  2003/04/23  Added monochrome bitmap generation.
 *  Version: 3.45  2003/05/05  Added some conversions.
 */

/* Copyright (c) L. Patrick

   This file is part of the App cross-platform programming package.
   You may redistribute it and/or modify it under the terms of the
   App Software License. See the file LICENSE.TXT for details.
*/

#include "appint.h"

/*
 *  Find a Pixmap format that will work:
 */

static void app_find_best_pixmap_format(Display *disp, int required_depth,
	int *actual_depth, int *bits_per_pixel, int *scanline_pad)
{
	int i, best, dist, count;
	XPixmapFormatValues *values;

	values = XListPixmapFormats(disp, &count);
	if (count == 0)
		return;
	best = 0;
	dist = 1000;

	for (i=0; i < count; i++) {
		if (values[i].depth == required_depth) {
			best = i;
			break;
		}
		else if (values[i].depth > required_depth) {
			if (values[i].depth - required_depth < dist) {
				dist = values[i].depth - required_depth;
				best = i;
			}
		}
	}

	*actual_depth = values[best].depth;
	*bits_per_pixel = values[best].bits_per_pixel;
	*scanline_pad = values[best].scanline_pad;
}

/*
 *  Functions which move data from an Image into an XImage scanline:
 */

static void app_row_8_1_1_MSB(byte *dst, byte *src, int width)
{
	/* copy 8-bit pixvals into 1 significant bit, each packed
	 * into 1 bits-per-pixel, first pixel in most significant spot */
	int i;

	for (i=0; i < width; i++,dst++)
	{
		*dst  = ((*src++ & 0x01)<<7); if (++i == width) break;
		*dst |= ((*src++ & 0x01)<<6); if (++i == width) break;
		*dst |= ((*src++ & 0x01)<<5); if (++i == width) break;
		*dst |= ((*src++ & 0x01)<<4); if (++i == width) break;
		*dst |= ((*src++ & 0x01)<<3); if (++i == width) break;
		*dst |= ((*src++ & 0x01)<<2); if (++i == width) break;
		*dst |= ((*src++ & 0x01)<<1); if (++i == width) break;
		*dst |=  (*src++ & 0x01);
	}
}

static void app_row_8_1_1_LSB(byte *dst, byte *src, int width)
{
	/* copy 8-bit pixvals into 1 significant bit, each packed
	 * into 1 bits-per-pixel, first pixel in least significant spot */
	int i;

	for (i=0; i < width; i++,dst++)
	{
		*dst  =  (*src++ & 0x01);     if (++i == width) break;
		*dst |= ((*src++ & 0x01)<<1); if (++i == width) break;
		*dst |= ((*src++ & 0x01)<<2); if (++i == width) break;
		*dst |= ((*src++ & 0x01)<<3); if (++i == width) break;
		*dst |= ((*src++ & 0x01)<<4); if (++i == width) break;
		*dst |= ((*src++ & 0x01)<<5); if (++i == width) break;
		*dst |= ((*src++ & 0x01)<<6); if (++i == width) break;
		*dst |= ((*src++ & 0x01)<<7);
	}
}

static void app_row_8_1_2_MSB(byte *dst, byte *src, int width)
{
	/* copy 8-bit pixvals into 1 significant bit, each packed
	 * into 2 bits-per-pixel, first pixel in most significant spot */
	int i;

	for (i=0; i < width; i++,dst++)
	{
		*dst  = ((*src++ & 0x01)<<6); if (++i == width) break;
		*dst |= ((*src++ & 0x01)<<4); if (++i == width) break;
		*dst |= ((*src++ & 0x01)<<2); if (++i == width) break;
		*dst |=  (*src++ & 0x01);
	}
}

static void app_row_8_1_2_LSB(byte *dst, byte *src, int width)
{
	/* copy 8-bit pixvals into 1 significant bit, each packed
	 * into 2 bits-per-pixel, first pixel in least significant spot */
	int i;

	for (i=0; i < width; i++,dst++)
	{
		*dst  =  (*src++ & 0x01);     if (++i == width) break;
		*dst |= ((*src++ & 0x01)<<2); if (++i == width) break;
		*dst |= ((*src++ & 0x01)<<4); if (++i == width) break;
		*dst |= ((*src++ & 0x01)<<6);
	}
}

static void app_row_8_1_4_MSB(byte *dst, byte *src, int width)
{
	/* copy 8-bit pixvals into 1 significant bit, each packed
	 * into 4 bits-per-pixel, first pixel in most significant spot */
	int i;

	for (i=0; i < width; i++,dst++)
	{
		*dst  = ((*src++ & 0x01)<<4); if (++i == width) break;
		*dst |=  (*src++ & 0x01);
	}
}

static void app_row_8_1_4_LSB(byte *dst, byte *src, int width)
{
	/* copy 8-bit pixvals into 1 significant bit, each packed
	 * into 4 bits-per-pixel, first pixel in least significant spot */
	int i;

	for (i=0; i < width; i++,dst++)
	{
		*dst  =  (*src++ & 0x01);     if (++i == width) break;
		*dst |= ((*src++ & 0x01)<<4);
	}
}

static void app_row_8_1_8(byte *dst, byte *src, int width)
{
	/* copy 8-bit pixvals into 1 significant bit, each packed
	 * into 8 bits-per-pixel */
	int i;

	for (i=0; i < width; i++,dst++)
	{
		*dst  =  (*src++ & 0x01);
	}
}

static void app_row_8_2_2_MSB(byte *dst, byte *src, int width)
{
	/* copy 8-bit pixvals into 2 significant bits, each packed
	 * into 2 bits-per-pixel, first pixel in most significant spot */
	int i;

	for (i=0; i < width; i++,dst++)
	{
		*dst  = ((*src++ & 0x03)<<6); if (++i == width) break;
		*dst |= ((*src++ & 0x03)<<4); if (++i == width) break;
		*dst |= ((*src++ & 0x03)<<2); if (++i == width) break;
		*dst |=  (*src++ & 0x03);
	}
}

static void app_row_8_2_2_LSB(byte *dst, byte *src, int width)
{
	/* copy 8-bit pixvals into 2 significant bits, each packed
	 * into 2 bits-per-pixel, first pixel in least significant spot */
	int i;

	for (i=0; i < width; i++,dst++)
	{
		*dst  =  (*src++ & 0x03);     if (++i == width) break;
		*dst |= ((*src++ & 0x03)<<2); if (++i == width) break;
		*dst |= ((*src++ & 0x03)<<4); if (++i == width) break;
		*dst |= ((*src++ & 0x03)<<6);
	}
}

static void app_row_8_2_4_MSB(byte *dst, byte *src, int width)
{
	/* copy 8-bit pixvals into 2 significant bits, each packed
	 * into 4 bits-per-pixel, first pixel in most significant spot */
	int i;

	for (i=0; i < width; i++,dst++)
	{
		*dst  = ((*src++ & 0x03)<<4); if (++i == width) break;
		*dst |=  (*src++ & 0x03);
	}
}

static void app_row_8_2_4_LSB(byte *dst, byte *src, int width)
{
	/* copy 8-bit pixvals into 2 significant bits, each packed
	 * into 4 bits-per-pixel, first pixel in least significant spot */
	int i;

	for (i=0; i < width; i++,dst++)
	{
		*dst  =  (*src++ & 0x03);     if (++i == width) break;
		*dst |= ((*src++ & 0x03)<<4);
	}
}

static void app_row_8_2_8(byte *dst, byte *src, int width)
{
	/* copy 8-bit pixvals into 2 significant bits, each packed
	 * into 8 bits-per-pixel */
	int i;

	for (i=0; i < width; i++,dst++)
	{
		*dst  =  (*src++ & 0x03);
	}
}

static void app_row_8_4_4_MSB(byte *dst, byte *src, int width)
{
	/* copy 8-bit pixvals into 4 significant bits, each packed
	 * into 4 bits-per-pixel, first pixel in most significant spot */
	int i;

	for (i=0; i < width; i++,dst++)
	{
		*dst  = ((*src++ & 0x0f)<<4); if (++i == width) break;
		*dst |=  (*src++ & 0x0f);
	}
}

static void app_row_8_4_4_LSB(byte *dst, byte *src, int width)
{
	/* copy 8-bit pixvals into 4 significant bits, each packed
	 * into 4 bits-per-pixel, first pixel in least significant spot */
	int i;

	for (i=0; i < width; i++,dst++)
	{
		*dst  =  (*src++ & 0x0f);     if (++i == width) break;
		*dst |= ((*src++ & 0x0f)<<4);
	}
}

static void app_row_8_4_8(byte *dst, byte *src, int width)
{
	/* copy 8-bit pixvals into 4 significant bits, each packed
	 * into 8 bits-per-pixel */
	int i;

	for (i=0; i < width; i++,dst++)
	{
		*dst  =  (*src++ & 0x0f);
	}
}

static void app_row_8_8_8(byte *dst, byte *src, int width)
{
	/* copy 8-bit pixvals into 8 significant bits, each packed
	 * into 8 bits-per-pixel */

	memcpy(dst, src, width);
}

static void app_row_12_16_rgb_MSB(byte *dst, Colour *src, int width)
{
	/* copy 12-bit RGBs into 16 bits-per-pixel, top byte first
	 * 4 bits per colour i.e. 0000rrrr ggggbbbb */
	int i;

	for (i=0; i < width; i++,dst++,src++)
	{
		*dst  = ((src->red>>4)  & 0x0f); dst++;
		*dst  = ((src->green)   & 0xf0);
		*dst |= ((src->blue>>4) & 0x0f);
	}
}

static void app_row_12_16_rgb_LSB(byte *dst, Colour *src, int width)
{
	/* copy 12-bit RGBs into 16 bits-per-pixel, top byte last
	 * 4 bits per colour i.e. ggggbbbb 0000rrrr */
	int i;

	for (i=0; i < width; i++,dst++,src++)
	{
		*dst  = ((src->green)   & 0xf0);
		*dst |= ((src->blue>>4) & 0x0f); dst++;
		*dst  = ((src->red>>4)  & 0x0f);
	}
}

static void app_row_15_16_rgb_MSB(byte *dst, Colour *src, int width)
{
	/* copy 15-bit RGBs into 16 bits-per-pixel, top byte first
	 * 5 bits per colour i.e. 0rrrrrgg gggbbbbb */
	int i;

	for (i=0; i < width; i++,dst++,src++)
	{
		*dst  = ((src->red>>1)   & 0x7c);
		*dst |= ((src->green>>6) & 0x03); dst++;
		*dst  = ((src->green<<2) & 0xe0);
		*dst |= ((src->blue>>3)  & 0x1f);
	}
}

static void app_row_15_16_rgb_LSB(byte *dst, Colour *src, int width)
{
	/* copy 15-bit RGBs into 16 bits-per-pixel, top byte last
	 * 5 bits per colour i.e. gggbbbbb 0rrrrrgg */
	int i;

	for (i=0; i < width; i++,dst++,src++)
	{
		*dst  = ((src->green<<2) & 0xe0);
		*dst |= ((src->blue>>3)  & 0x1f); dst++;
		*dst  = ((src->red>>1)   & 0x7c);
		*dst |= ((src->green>>6) & 0x03);
	}
}

static void app_row_16_rggb_MSB(byte *dst, Colour *src, int width)
{
	/* copy 16-bit RGBs into 16 bits-per-pixel, top byte first
	 * 5 bits per colour but 6 for green, i.e. rrrrrggg gggbbbbb */
	int i;

	for (i=0; i < width; i++,dst++,src++)
	{
		*dst  = ((src->red)      & 0xf8);
		*dst |= ((src->green>>5) & 0x07); dst++;
		*dst  = ((src->green<<3) & 0xe0);
		*dst |= ((src->blue>>3)  & 0x1f);
	}
}

static void app_row_16_rggb_LSB(byte *dst, Colour *src, int width)
{
	/* copy 16-bit RGBs into 16 bits-per-pixel, top byte last
	 * 5 bits per colour but 6 for green, i.e. gggbbbbb rrrrrggg */
	int i;

	for (i=0; i < width; i++,dst++,src++)
	{
		*dst  = ((src->green<<3) & 0xe0);
		*dst |= ((src->blue>>3)  & 0x1f); dst++;
		*dst  = ((src->red)      & 0xf8);
		*dst |= ((src->green>>5) & 0x07);
	}
}

static void app_row_24_rgb_MSB(byte *dst, Colour *src, int width)
{
	/* copy 24-bit RGBs into 24 bits-per-pixel, top byte first
	 * 8 bits per colour, i.e. rrrrrrrr gggggggg bbbbbbbb */
	int i;

	for (i=0; i < width; i++,dst++,src++)
	{
		*dst  = (src->red);   dst++;
		*dst  = (src->green); dst++;
		*dst  = (src->blue);
	}
}

static void app_row_24_rgb_LSB(byte *dst, Colour *src, int width)
{
	/* copy 24-bit RGBs into 24 bits-per-pixel, top byte last
	 * 8 bits per colour, i.e. bbbbbbbb gggggggg rrrrrrrr */
	int i;

	for (i=0; i < width; i++,dst++,src++)
	{
		*dst  = (src->blue);  dst++;
		*dst  = (src->green); dst++;
		*dst  = (src->red);
	}
}

static void app_row_24_32_rgb_MSB(byte *dst, Colour *src, int width)
{
	/* copy 24-bit RGBs into 32 bits-per-pixel, top byte first
	 * 8 bits per colour, i.e. aaaaaaaa rrrrrrrr gggggggg bbbbbbbb */
	int i;

	for (i=0; i < width; i++,dst++,src++)
	{
		*dst  = (src->alpha); dst++;
		*dst  = (src->red);   dst++;
		*dst  = (src->green); dst++;
		*dst  = (src->blue);
	}
}

static void app_row_24_32_rgb_LSB(byte *dst, Colour *src, int width)
{
	/* copy 24-bit RGBs into 32 bits-per-pixel, top byte last
	 * 8 bits per colour, i.e. bbbbbbbb gggggggg rrrrrrrr aaaaaaaa */
	int i;

	for (i=0; i < width; i++,dst++,src++)
	{
		*dst  = (src->blue);  dst++;
		*dst  = (src->green); dst++;
		*dst  = (src->red);   dst++;
		*dst  = (src->alpha);
	}
}

/*
 *  Declare a data structure which 'knows' about the above
 *  fast blitting functions and allows quick searching for
 *  an appropriate function to use:
 */

typedef void (*AppBltFunc)(byte *dst, void *src, int width);

typedef struct {
	int            depth, bpp, order, use_masks;
	unsigned long  rmask, gmask, bmask;
	void *         fn;
} AppBltFuncStruct;

static AppBltFuncStruct app_blt_func_table[] =
{
 /* 8-bit pixvals to varying depth colour maps */

 { 1, 1,1,0,0UL,0UL,0UL,app_row_8_1_1_MSB},
 { 1, 1,0,0,0UL,0UL,0UL,app_row_8_1_1_LSB},

 { 1, 2,1,0,0UL,0UL,0UL,app_row_8_1_2_MSB},
 { 1, 2,0,0,0UL,0UL,0UL,app_row_8_1_2_LSB},

 { 1, 4,1,0,0UL,0UL,0UL,app_row_8_1_4_MSB},
 { 1, 4,0,0,0UL,0UL,0UL,app_row_8_1_4_LSB},

 { 1, 8,1,0,0UL,0UL,0UL,app_row_8_1_8},
 { 1, 8,0,0,0UL,0UL,0UL,app_row_8_1_8},

 { 2, 2,1,0,0UL,0UL,0UL,app_row_8_2_2_MSB},
 { 2, 2,0,0,0UL,0UL,0UL,app_row_8_2_2_LSB},

 { 2, 4,1,0,0UL,0UL,0UL,app_row_8_2_4_MSB},
 { 2, 4,0,0,0UL,0UL,0UL,app_row_8_2_4_LSB},

 { 2, 8,1,0,0UL,0UL,0UL,app_row_8_2_8},
 { 2, 8,0,0,0UL,0UL,0UL,app_row_8_2_8},

 { 4, 4,1,0,0UL,0UL,0UL,app_row_8_4_4_MSB},
 { 4, 4,0,0,0UL,0UL,0UL,app_row_8_4_4_LSB},

 { 4, 8,1,0,0UL,0UL,0UL,app_row_8_4_8},
 { 4, 8,0,0,0UL,0UL,0UL,app_row_8_4_8},

 { 8, 8,1,0,0UL,0UL,0UL,app_row_8_8_8},
 { 8, 8,0,0,0UL,0UL,0UL,app_row_8_8_8},

 /* 32-bit rgbs to true colour pixels using component masks */

 {12,16,1,1,0x0000f00UL,0x00000f0UL,0x000000fUL,app_row_12_16_rgb_MSB},
 {12,16,0,1,0x0000f00UL,0x00000f0UL,0x000000fUL,app_row_12_16_rgb_LSB},

 {15,16,1,1,0x0007c00UL,0x00003e0UL,0x000001fUL,app_row_15_16_rgb_MSB},
 {15,16,0,1,0x0007c00UL,0x00003e0UL,0x000001fUL,app_row_15_16_rgb_LSB},

 {16,16,1,1,0x000f800UL,0x00007e0UL,0x000001fUL,app_row_16_rggb_MSB},
 {16,16,0,1,0x000f800UL,0x00007e0UL,0x000001fUL,app_row_16_rggb_LSB},

 {24,24,1,1,0x0ff0000UL,0x000ff00UL,0x00000ffUL,app_row_24_rgb_MSB},
 {24,24,0,1,0x0ff0000UL,0x000ff00UL,0x00000ffUL,app_row_24_rgb_LSB},

 {24,32,1,1,0x0ff0000UL,0x000ff00UL,0x00000ffUL,app_row_24_32_rgb_MSB},
 {24,32,0,1,0x0ff0000UL,0x000ff00UL,0x00000ffUL,app_row_24_32_rgb_LSB},

 /* converse cases */

 {12,16,1,1,0x000000fUL,0x00000f0UL,0x0000f00UL,app_row_12_16_rgb_MSB},
 {12,16,0,1,0x000000fUL,0x00000f0UL,0x0000f00UL,app_row_12_16_rgb_LSB},

 {15,16,1,1,0x000001fUL,0x00003e0UL,0x0007c00UL,app_row_15_16_rgb_MSB},
 {15,16,0,1,0x000001fUL,0x00003e0UL,0x0007c00UL,app_row_15_16_rgb_LSB},

 {16,16,1,1,0x000001fUL,0x00007e0UL,0x000f800UL,app_row_16_rggb_MSB},
 {16,16,0,1,0x000001fUL,0x00007e0UL,0x000f800UL,app_row_16_rggb_LSB},

 {24,24,1,1,0x00000ffUL,0x000ff00UL,0x0ff0000UL,app_row_24_rgb_MSB},
 {24,24,0,1,0x00000ffUL,0x000ff00UL,0x0ff0000UL,app_row_24_rgb_LSB},

 {24,32,1,1,0x00000ffUL,0x000ff00UL,0x0ff0000UL,app_row_24_32_rgb_MSB},
 {24,32,0,1,0x00000ffUL,0x000ff00UL,0x0ff0000UL,app_row_24_32_rgb_LSB}
};

static AppBltFunc app_find_blt_func(int depth,
	int bits_per_pixel, int order, int use_masks,
	unsigned long red_mask, unsigned long green_mask,
	unsigned long blue_mask)
{
	int i, nelem;

	nelem = sizeof(app_blt_func_table)/sizeof(app_blt_func_table[0]);

	for (i=0; i < nelem; i++) {
		if ((depth == app_blt_func_table[i].depth) &&
		    (bits_per_pixel == app_blt_func_table[i].bpp) &&
		    (order == app_blt_func_table[i].order) &&
		    (use_masks == app_blt_func_table[i].use_masks))
		{
			if (! use_masks) {
				return app_blt_func_table[i].fn;
			}
			else if
			   ((red_mask   == app_blt_func_table[i].rmask) &&
			    (green_mask == app_blt_func_table[i].gmask) &&
			    (blue_mask  == app_blt_func_table[i].bmask))
			{
				return app_blt_func_table[i].fn;
			}
		}
	}
	return NULL;	/* no matching row function found */
}

/*
 *  Determine if a display is using TrueColor or DirectColor.
 *  A TrueColor display uses the pixel value to directly determine
 *  the RGB intensities (e.g. 24-bit displays). DirectColor is
 *  similar, but the actual palette of colours is variable.
 *  Return zero if the display is using a colormap.
 */
int app_is_true_colour_display(Display *disp)
{
	Visual *vis;
	int vis_class;

	vis = DefaultVisual(disp, DefaultScreen(disp));
	if (! vis)
		return 0;

	#if defined(__cplusplus) || defined(c_plusplus)
         vis_class = vis->c_class;
	#else
         vis_class = vis->class;
	#endif

	if ((vis_class == TrueColor) || (vis_class == DirectColor))
		return 1;
	return 0;
}

/*
 *  If we cannot find an appropriate fast row blitting function,
 *  don't give up! We can use this slow, but correct, method
 *  of drawing the bits:
 */
static void app_draw_image_slowly(Display *disp, XID xid, GC gc,
	Graphics *g, int width, int height, Image *img)
{
	int x, y;
	int value;
	Colour col;

	if (img->depth == 8) {
	  for (y=0; y < height; y++) {
		for (x=0; x < width; x++) {
			value = img->data8[y][x];
			col = img->cmap[value];
			app_set_rgb(g, col);
			XFillRectangle(disp, xid, gc, x, y, 1, 1);
		}
	  }
	}
	else if (img->depth == 32) {
	  for (y=0; y < height; y++) {
		for (x=0; x < width; x++) {
			col = img->data32[y][x];
			app_set_rgb(g, col);
			XFillRectangle(disp, xid, gc, x, y, 1, 1);
		}
	  }
	}
}

/*
 *  Return a 'clipmask' Pixmap. This is a Pixmap of depth 1,
 *  with 1's where the image is opaque, 0's where it is transparent.
 *  We just make it completely transparent here.
 *  Return None on failure.
 */
XID app_new_clipmask(Display *disp, int width, int height)
{
	Pixmap bmap;
	GC gc;

	bmap = XCreatePixmap(disp, DefaultRootWindow(disp), width, height, 1);

	if (! app_bitmap_created(disp, bmap))
		return None;

	gc = XCreateGC(disp, bmap, 0, NULL);
	if (! gc) {
		XFreePixmap(disp, bmap);
		return None;
	}

	XSetForeground(disp, gc, 0); /* fill with zeros => transparent */
	XFillRectangle(disp, bmap, gc, 0, 0, width, height);
	XFreeGC(disp, gc);

	return bmap;
}

/*
 *  Copy transparent pixels to 0-bits, opaque to 1-bits.
 */
static void app_row_8_T_1(byte *dst, byte *src, int width, Colour *cmap,
	int order)
{
	int x, shift, value;
	Colour col;

	shift = 7;
	for (x=0; x < width; x++) {
		value = src[x];
		col = cmap[value];
		if (col.alpha <= 0x7F) {
			if (order == MSBFirst)
				*dst |= (1<<shift);
			else
				*dst |= (1<<(7-shift));
		}
		if (--shift < 0) {
			shift = 7;
			dst++;
		}
	}
}

static void app_row_32_T_1(byte *dst, Colour *src, int width, int order)
{
	int x, shift;
	Colour col;

	shift = 7;
	for (x=0; x < width; x++) {
		col = src[x];
		if (col.alpha <= 0x7F) {
			if (order == MSBFirst)
				*dst |= (1<<shift);
			else
				*dst |= (1<<(7-shift));
		}
		if (--shift < 0) {
			shift = 7;
			dst++;
		}
	}
}

/*
 *  Create a clipmask Pixmap.
 *
 *  This creates a clipmask that has 0's wherever the source
 *  image is transparent, 1's elsewhere. It checks that there is at
 *  least one transparent pixel in the image first, and if not,
 *  returns None which is equivalent to having no clipping.
 */

XID app_image_to_clipmask(App *app, Image *img)
{
	int width, height, depth, rowbytes;
	int y, bit_order;
	int bits_per_pixel, scanline_pad;
	byte *data;
	XImage *xi;
	Display *disp;
	GC gc;
	Pixmap clipmask;

	if (! img)
		return None;
	if ((img->depth != 8) && (img->depth != 32))
		return None;
	if (! app_image_has_transparent_pixels(img))
		return None;

	disp = app_extra(app)->display;

	width = img->width;
	height = img->height;

	app_find_best_pixmap_format(disp, 1, &depth,
		&bits_per_pixel, &scanline_pad);

	if ((depth != 1) || (bits_per_pixel != 1))
		return None;

	xi = XCreateImage(disp,
		DefaultVisual(disp, DefaultScreen(disp)),
		1, ZPixmap, 0, NULL, width, height, scanline_pad, 0);
	if (! xi) 
		return None;

	rowbytes = xi->bytes_per_line;
	bit_order = xi->bitmap_bit_order;

	data = app_zero_alloc(height * rowbytes);
	if (! data) {
		XDestroyImage(xi);
		return None;
	}

	if (img->depth == 8)
	{
		/* assign pixels into data array */
		for (y=0; y < height; y++)
			app_row_8_T_1(&data[y*rowbytes],
				img->data8[y], width,
				img->cmap, bit_order);
	}
	else if (img->depth == 32)
	{
		/* assign pixels into data array */
		for (y=0; y < height; y++)
			app_row_32_T_1(&data[y*rowbytes],
				img->data32[y], width, bit_order);
	}

	clipmask = app_new_clipmask(disp, width, height);
	if (clipmask == None) {
		app_free(data);
		XDestroyImage(xi);
		return None;
	}

	gc = XCreateGC(disp, clipmask, 0, NULL);
	if (! gc) {
		XFreePixmap(disp, clipmask);
		app_free(data);
		XDestroyImage(xi);
		return None;
	}

	xi->data = data;
	XPutImage(disp, clipmask, gc, xi, 0,0,0,0, width, height);
	xi->data = NULL;

	XFreeGC(disp, gc);
	app_free(data);
	XDestroyImage(xi);

	return clipmask;
}

/*
 *  Create a bitmap from an image.
 *  The bitmap will be a Pixmap with the bits of the image,
 *  and there may also be a clipmask Pixmap if the image
 *  contains any transparent pixels.
 */
enum RenderingTechnique {
	SlowRender               = 0,
	Fast8BitToSystemPalette  = 1,
	Fast32BitToSystemPalette = 2,
	Fast8BitToWindowPalette  = 3,
	Fast32BitToWindowPalette = 4,
	Fast8BitToDirectColour   = 5,
	Fast32BitToDirectColour  = 6
};

Bitmap *app_image_to_bitmap(Window *win, Image *img)
{
	int width, height, depth, rowbytes;
	int x, y, value;
	int bits_per_pixel, scanline_pad, order;
	int technique;
	AppBltFunc blt_func;
	Display *disp;
	XImage *xi;
	Bitmap *b;
	Graphics *g;
	byte *data = NULL;
	byte *translation = NULL;
	byte *byte_row = NULL;
	Colour *rgb_row = NULL;

	technique = SlowRender;
	disp = app_extra(win->app)->display;
	width = img->width;
	height = img->height;

	b = app_new_bitmap(win, width, height);
	if (! b) {
		/* could not create the bitmap! */
		return NULL;
	}

	g = app_get_bitmap_graphics(b);
	if (! g) {
		/* cannot draw the image if we can't get a GC! */
		app_del_bitmap(b);
		return NULL;
	}

	rowbytes = 0;
	depth = DefaultDepth(disp, DefaultScreen(disp));
	app_find_best_pixmap_format(disp, depth, &depth,
		&bits_per_pixel, &scanline_pad);

	xi = XCreateImage(disp,
		DefaultVisual(disp, DefaultScreen(disp)), depth,
		ZPixmap, 0, NULL, width, height, scanline_pad, 0);

	if (xi) {
		/* success */
		rowbytes = xi->bytes_per_line;
		depth = xi->depth;
		bits_per_pixel = xi->bits_per_pixel;

		if (depth <= 8)
			order = ((xi->bitmap_bit_order==MSBFirst) ?1:0);
		else
			order = ((xi->byte_order==MSBFirst) ?1:0);

		if (win_extra(win)->is_paletted)
			blt_func = app_find_blt_func(depth,
				bits_per_pixel, order,
				0, 0UL, 0UL, 0UL);
		else
			blt_func = app_find_blt_func(depth,
				bits_per_pixel, order, 1, xi->red_mask,
				xi->green_mask, xi->blue_mask);

		if (blt_func) {
			data = app_alloc(height * rowbytes);
			if (! data)
				technique = SlowRender;
		}
	}
	else {
		/* failed to allocate the XImage */
		blt_func = NULL;
	}

	if (blt_func == NULL) {
		/* could not find a fast row blitting function */
		/* fall back to a slow, correct drawing method */
		technique = SlowRender;
	}
	else if ((img->depth == 8) && (! win_extra(win)->is_paletted))
	{
		technique = Fast8BitToDirectColour;
		rgb_row = app_alloc(width * sizeof(Colour));
		if (! rgb_row)
			technique = SlowRender;
	}
	else if ((img->depth == 32) && (! win_extra(win)->is_paletted))
	{
		technique = Fast32BitToDirectColour;
	}
	else if ((img->depth == 8) && (win->pal))
	{
		technique = Fast8BitToWindowPalette;
		translation = app_alloc(img->cmap_size);
		byte_row = app_alloc(width);
		if ((! translation) || (! byte_row))
			technique = SlowRender;
	}
	else if ((img->depth == 32) && (win->pal))
	{
		technique = Fast32BitToWindowPalette;
		byte_row = app_alloc(width);
		if (! byte_row)
			technique = SlowRender;
	}
	else if (img->depth == 8)
	{
		technique = Fast8BitToSystemPalette;
		if (! app_extra(win->app)->pal)
			technique = SlowRender;
		else {
			translation = app_alloc(img->cmap_size);
			byte_row = app_alloc(width);
			if ((! translation) || (! byte_row))
				technique = SlowRender;
		}
	}
	else if (img->depth == 32)
	{
		technique = Fast32BitToSystemPalette;
		if (! app_extra(win->app)->pal)
			technique = SlowRender;
		else {
			byte_row = app_alloc(width);
			if (! byte_row)
				technique = SlowRender;
		}
	}
	else {
		technique = SlowRender;
	}

	/* now render using the correct technique */

	if (technique == Fast8BitToDirectColour)
	{
		/* expand image's pixvals to image's cmap colours */
		/* then blit from rgb_row array into data array */
		for (y=0; y < height; y++) {
			for (x=0; x < width; x++) {
				value = img->data8[y][x];
				rgb_row[x] = img->cmap[value];
			}
			blt_func(&data[y*rowbytes], rgb_row, width);
		}
	}
	else if (technique == Fast32BitToDirectColour)
	{
		/* just call the row blitting function on each row */
		for (y=0; y < height; y++) {
			blt_func(&data[y*rowbytes], img->data32[y], width);
		}
	}
	else if (technique == Fast8BitToWindowPalette)
	{
		/* translate img->cmap to win->pal once */
		/* then use it to translate each row's 8-bit values */

		app_palette_translation(win->pal, translation,
			img->cmap_size, img->cmap);

		/* assign pixels into data array */
		for (y=0; y < height; y++) {
			for (x=0; x < width; x++) {
				value = img->data8[y][x];
				byte_row[x] = translation[value];
			}
			blt_func(&data[y*rowbytes], byte_row, width);
		}
	}
	else if (technique == Fast32BitToWindowPalette)
	{
		/* translate each row of rgbs to win->pal */
		for (y=0; y < height; y++) {
			app_palette_translation(win->pal, byte_row,
				width, img->data32[y]);
			blt_func(&data[y*rowbytes], byte_row, width);
		}
	}
	else if (technique == Fast8BitToSystemPalette)
	{
		/* translate from img->cmap to win->app->pal once */
		/* then use it to translate each row's 8-bit values */

		app_palette_translation(app_extra(win->app)->pal,
			translation, img->cmap_size, img->cmap);

		/* assign pixels into data array */
		for (y=0; y < height; y++) {
			for (x=0; x < width; x++) {
				value = img->data8[y][x];
				byte_row[x] = translation[value];
			}
			blt_func(&data[y*rowbytes], byte_row, width);
		}
	}
	else if (technique == Fast32BitToSystemPalette)
	{
		/* translate each row of rgbs to win->app->pal */
		for (y=0; y < height; y++) {
			app_palette_translation(app_extra(win->app)->pal,
				byte_row, width, img->data32[y]);

			blt_func(&data[y*rowbytes], byte_row, width);
		}
	}
	else {
		/* just use the slow way of drawing */
		app_draw_image_slowly(disp, bitmap_extra(b)->handle,
			graphics_extra(g)->gc, g, width, height, img);
	}

	/* discard temporary data structures */

	if (byte_row)
		app_free(byte_row);
	if (rgb_row)
		app_free(rgb_row);
	if (translation)
		app_free(translation);

	/* copy bits over, if we haven't already done it the slow way */

	if (technique != SlowRender) {
		if (xi) {
			if (data) {
				xi->data = data;
				XPutImage(disp, bitmap_extra(b)->handle,
					graphics_extra(g)->gc, xi,
					0,0,0,0, width, height);
				xi->data = NULL;
			}
			XDestroyImage(xi);
		}
	}

	/* clean up and produce the transparency mask, if needed */

	if (data)
		app_free(data);
	app_del_graphics(g);
	if (bitmap_extra(b)->clipmask != None)
		XFreePixmap(app_extra(win->app)->display,
			bitmap_extra(b)->clipmask);
	bitmap_extra(b)->clipmask = app_image_to_clipmask(win->app, img);
	return b;
}

/*
 *  Generate monochrome (black and white) Bitmaps:
 *  This code generates a Bitmap with depth 1 and a clipmask with depth 1.
 *  The Bitmap is generated using a slow but correct algorithm,
 *  which accumulates adjacent colours and draws each row in pieces.
 *  The clipmask is always present even if the image is fully opaque.
 */
static Colour app_monochrome(Colour c)
{
	int min, max, g, b;

	max = min = c.red;
	g = c.green;
	if      (g < min) min = g;
	else if (g > max) max = g;
	b = c.blue;
	if      (b < min) min = b;
	else if (b > max) max = b;

	if (min > 0xE0) 	c.red = c.green = c.blue = 255;
	else if (max < 0x10)	c.red = c.green = c.blue = 0;
	else if (max < 0x60)	c.red = c.green = c.blue = 0;
	else if (max < 0xD0)	c.red = c.green = c.blue = 0;
	else			c.red = c.green = c.blue = 255;

	return c;
}

static Colour app_get_monochrome_pixel_at(Image *img, int x, int y)
{
	if (img->depth == 8)
		return app_monochrome(img->cmap[img->data8[y][x]]);
	else
		return app_monochrome(img->data32[y][x]);
}

Bitmap *app_image_to_monochrome_bitmap(Window *win, Image *img)
{
	Bitmap *bmp;
	int x, y, xstart, xwidth;
	Colour col, prev;
	Display *disp;
	XID dst_id, dst_mask;
	GC dst_gc, dst_mask_gc = 0;

	bmp = app_new_monochrome_bitmap(win, img->width, img->height);
	if (bmp == NULL)
		return NULL;

	disp = app_extra(bmp->win->app)->display;
	dst_id = bitmap_extra(bmp)->handle;
	dst_mask = bitmap_extra(bmp)->clipmask;
	dst_gc = XCreateGC(disp, dst_id, 0, NULL);
	if (! dst_gc) {
		app_del_bitmap(bmp);
		return NULL;
	}
	dst_mask_gc = XCreateGC(disp, dst_mask, 0, NULL);
	if (! dst_mask_gc) {
		XFreeGC(disp, dst_gc);
		app_del_bitmap(bmp);
		return NULL;
	}
	XSetForeground(disp, dst_gc, 0);
	XFillRectangle(disp, dst_id, dst_gc, 0, 0, img->width, img->height);
	XSetForeground(disp, dst_gc, ~0);
	XSetForeground(disp, dst_mask_gc, ~0);

	for (y=0; y < img->height; y++) {
		for (x=xstart=xwidth=0; x <= img->width; x++) {
			if (x < img->width)
				col = app_get_monochrome_pixel_at(img,x,y);
			if (x == img->width) {
				/* fall through to output the row */
			}
			else if (xwidth == 0) {
				/* start output row with this pixel */
				xwidth = 1;
				prev = col;
				continue;
			}
			else if (app_colours_equal(prev, col)) {
				/* add this pixel to the output row */
				xwidth++;
				continue;
			}
			/* output the row since this colour differs */
			if (prev.blue <= 0x7F)
				XFillRectangle(disp, dst_id, dst_gc,
						xstart, y, xwidth, 1);
			if (prev.alpha <= 0x7F)
				XFillRectangle(disp, dst_mask, dst_mask_gc,
						xstart, y, xwidth, 1);
			xstart = x;
			xwidth = 1;
			prev = col;
		}
	}
	XFreeGC(disp, dst_mask_gc);
	XFreeGC(disp, dst_gc);

	return bmp;
}
